def func_best_model(scan_object, metric='val_acc', asc=False):

    '''Picks the best model based on a given metric and
    returns the index number for the model.

    NOTE: for loss 'asc' should be True'''

    import warnings as warnings

    warnings.simplefilter('ignore')

    from ..utils.best_model import best_model, activate_model
    model_no = best_model(scan_object, metric, asc)
    out = activate_model(scan_object, model_no)

    return out


def func_evaluate(scan_object,
                  x_val,
                  y_val,
                  task,
                  n_models=10,
                  metric='val_acc',
                  folds=5,
                  shuffle=True,
                  asc=False):

    '''K-fold Cross Evaluator

    For creating scores from kfold cross-evaluation and
    adding them to the data frame.

    scan_object : python class
        The class object returned by Scan() upon completion of the experiment.
    x_val : array or list of arrays
        Input data (features) in the same format as used in Scan(), but should
        not be the same data (or it will not be much of validation).
    y_val : array or list of arrays
        Input data (labels) in the same format as used in Scan(), but should
        not be the same data (or it will not be much of validation).
    task : string
        'binary', 'multi_class', 'multi_label', or 'continuous'.
    n_models : int
        The number of models to be evaluated. If set to 10, then 10 models
        with the highest metric value are evaluated. See below.
    metric : str
        The metric to be used for picking the models to be evaluated.
    folds : int
        The number of folds to be used in the evaluation.
    shuffle : bool
        If the data is to be shuffled or not. Set always to False for
        timeseries but keep in mind that you might get periodical/seasonal bias.
    asc : bool
        Set to True if the metric is to be minimized.

    '''
    import warnings as warnings
    from tqdm import tqdm
    import numpy as np

    warnings.simplefilter('ignore')

    picks = scan_object.data.sort_values(metric,
                                         ascending=asc).index.values[:n_models]

    if n_models > len(scan_object.data):
        data_len = len(scan_object.data)
    else:
        data_len = n_models

    out = []

    pbar = tqdm(total=data_len)

    from ..commands.evaluate import Evaluate
    for i in range(len(scan_object.data)):
        if i in list(picks):
            evaluate_object = Evaluate(scan_object)
            temp = evaluate_object.evaluate(x_val, y_val,
                                            task=task,
                                            model_id=i,
                                            metric=metric,
                                            folds=folds,
                                            shuffle=shuffle,
                                            asc=asc)
            out.append([np.mean(temp), np.std(temp)])
            pbar.update(1)
        else:
            out.append([np.nan, np.nan])

    pbar.close()

    scan_object.data['eval_f1score_mean'] = [i[0] for i in out]
    scan_object.data['eval_f1score_std'] = [i[1] for i in out]

    print(">> Added evaluation score columns to scan_object.data")



import warnings
warnings.simplefilter('ignore')

# import commands
from .scan.Scan import Scan
from .commands.analyze import Analyze
from .commands.analyze import Analyze as Reporting
from .commands.predict import Predict
from .commands.deploy import Deploy
from .commands.evaluate import Evaluate
from .commands.restore import Restore

# import extras
from . import utils
from . import templates
from . import autom8

# the purpose of everything below is to keep the namespace completely clean

template_sub = [templates.datasets,
                templates.models,
                templates.params,
                templates.pipelines]

keep_from_templates = ['iris', 'cervical_cancer', 'titanic', 'breast_cancer',
                       'icu_mortality', 'telco_churn', 'mnist']

for sub in template_sub:
    for key in list(sub.__dict__):
        if key.startswith('__') is False:
            if key not in keep_from_templates:
                delattr(sub, key)

del commands, scan, model, metrics, key
del sub, keep_from_templates, template_sub, warnings

__version__ = "0.6.3"



def scan_round(self):

    '''The main operational function that manages the experiment
    on the level of execution of each round.'''

    import time
    import gc

    # print round params
    if self.print_params is True:
        print(self.round_params)

    # set start time
    round_start = time.strftime('%D-%H%M%S')
    start = time.time()

    # fit the model
    from ..model.ingest_model import ingest_model
    self.model_history, self.round_model = ingest_model(self)
    self.round_history.append(self.model_history.history)

    # handle logging of results
    from ..logging.logging_run import logging_run
    self = logging_run(self, round_start, start, self.model_history)

    # apply reductions
    from ..reducers.reduce_run import reduce_run
    self = reduce_run(self)

    try:
        # save model and weights
        self.saved_models.append(self.round_model.to_json())

        if self.save_weights:
            self.saved_weights.append(self.round_model.get_weights())
        else:
            self.saved_weights.append(None)

    except AttributeError as e:
        # make sure that the error message is from torch
        if str(e) == "'Model' object has no attribute 'to_json'":
            if self.save_weights:
                self.saved_models.append(self.round_model.state_dict())
            else:
                self.saved_weights.append(None)

    # clear tensorflow sessions
    if self.clear_session is True:

        del self.round_model
        gc.collect()

        # try TF specific and pass for everyone else
        try:
            from keras import backend as K
            K.clear_session()
        except:
            pass

    return self



def titanic():

    # here use a standard 2d dictionary for inputting the param boundaries
    p = {'lr': (0.5, 5, 10),
         'first_neuron': [4, 8, 16],
         'batch_size': [20, 30, 40],
         'dropout': (0, 0.5, 5),
         'optimizer': ['Adam', 'Nadam'],
         'losses': ['logcosh', 'binary_crossentropy'],
         'activation': ['relu', 'elu'],
         'last_activation': ['sigmoid']}

    return p


def iris():

    from keras.optimizers import Adam, Nadam
    from keras.losses import logcosh, categorical_crossentropy
    from keras.activations import relu, elu, softmax

    # here use a standard 2d dictionary for inputting the param boundaries
    p = {'lr': (0.5, 5, 10),
         'first_neuron': [4, 8, 16, 32, 64],
         'hidden_layers': [0, 1, 2, 3, 4],
         'batch_size': (2, 30, 10),
         'epochs': [2],
         'dropout': (0, 0.5, 5),
         'weight_regulizer': [None],
         'emb_output_dims':  [None],
         'shapes': ['brick', 'triangle', 0.2],
         'optimizer': [Adam, Nadam],
         'losses': [logcosh, categorical_crossentropy],
         'activation': [relu, elu],
         'last_activation': [softmax]}

    return p


def breast_cancer():

    from keras.optimizers import Adam, Nadam, RMSprop
    from keras.losses import logcosh, binary_crossentropy
    from keras.activations import relu, elu, sigmoid

    # then we can go ahead and set the parameter space
    p = {'lr': (0.5, 5, 10),
         'first_neuron': [4, 8, 16, 32, 64],
         'hidden_layers': [0, 1, 2],
         'batch_size': (2, 30, 10),
         'epochs': [50, 100, 150],
         'dropout': (0, 0.5, 5),
         'shapes': ['brick', 'triangle', 'funnel'],
         'optimizer': [Adam, Nadam, RMSprop],
         'losses': [logcosh, binary_crossentropy],
         'activation': [relu, elu],
         'last_activation': [sigmoid]}

    return p


def cervical_cancer():
    return breast_cancer()



def parallel_gpu_jobs(allow_growth=True, fraction=.5):

    '''Sets the max used memory as a fraction for tensorflow
    backend

    allow_growth :: True of False

    fraction :: a float value (e.g. 0.5 means 4gb out of 8gb)

    '''

    import keras.backend as K
    import tensorflow as tf

    gpu_options = tf.GPUOptions(allow_growth=allow_growth,
                                  per_process_gpu_memory_fraction=fraction)
    config = tf.ConfigProto(gpu_options=gpu_options)
    session = tf.Session(config=config)
    K.set_session(session)


def multi_gpu(model, gpus=None, cpu_merge=True, cpu_relocation=False):

    '''Takes as input the model, and returns a model
    based on the number of GPUs available on the machine
    or alternatively the 'gpus' user input.

    NOTE: this needs to be used before model.compile() in the
    model inputted to Scan in the form:

    from talos.utils.gpu_utils import multi_gpu
    model = multi_gpu(model)

    '''

    from keras.utils import multi_gpu_model

    return multi_gpu_model(model,
                           gpus=gpus,
                           cpu_merge=cpu_merge,
                           cpu_relocation=cpu_relocation)


def force_cpu():

    '''Force CPU on a GPU system
    '''

    import keras.backend as K
    import tensorflow as tf

    config = tf.ConfigProto(device_count={'GPU': 0})
    session = tf.Session(config=config)
    K.set_session(session)



def breast_cancer(x_train, y_train, x_val, y_val, params):

    from keras.models import Sequential
    from keras.layers import Dropout, Dense
    from talos.model import lr_normalizer, early_stopper, hidden_layers

    from talos.metrics.keras_metrics import matthews, precision, recall, f1score

    model = Sequential()
    model.add(Dense(params['first_neuron'],
                    input_dim=x_train.shape[1],
                    activation='relu'))

    model.add(Dropout(params['dropout']))

    hidden_layers(model, params, 1)

    model.add(Dense(1, activation=params['last_activation']))

    model.compile(optimizer=params['optimizer']
                  (lr=lr_normalizer(params['lr'],
                                    params['optimizer'])),
                  loss=params['losses'],
                  metrics=['acc',
                           f1score,
                           recall,
                           precision,
                           matthews])

    results = model.fit(x_train, y_train,
                        batch_size=params['batch_size'],
                        epochs=params['epochs'],
                        verbose=0,
                        validation_data=[x_val, y_val],
                        callbacks=[early_stopper(params['epochs'],
                                                 mode='moderate',
                                                 monitor='val_f1score')])

    return results, model


def cervical_cancer(x_train, y_train, x_val, y_val, params):

    from keras.models import Sequential
    from keras.layers import Dropout, Dense
    from talos.model import lr_normalizer, early_stopper, hidden_layers

    from talos.metrics.keras_metrics import matthews, precision, recall, f1score

    model = Sequential()
    model.add(Dense(params['first_neuron'],
                    input_dim=x_train.shape[1],
                    activation='relu'))

    model.add(Dropout(params['dropout']))

    hidden_layers(model, params, 1)

    model.add(Dense(1, activation=params['last_activation']))

    model.compile(optimizer=params['optimizer']
                  (lr=lr_normalizer(params['lr'],
                                    params['optimizer'])),
                  loss=params['losses'],
                  metrics=['acc',
                           f1score,
                           recall,
                           precision,
                           matthews])

    results = model.fit(x_train, y_train,
                        batch_size=params['batch_size'],
                        epochs=params['epochs'],
                        verbose=0,
                        validation_data=[x_val, y_val],
                        callbacks=[early_stopper(params['epochs'],
                                                 mode='moderate',
                                                 monitor='val_f1score')])

    return results, model


def titanic(x_train, y_train, x_val, y_val, params):

    from keras.models import Sequential
    from keras.layers import Dropout, Dense

    # note how instead of passing the value, we pass a dictionary entry
    model = Sequential()
    model.add(Dense(params['first_neuron'],
                    input_dim=x_train.shape[1],
                    activation='relu'))

    # same here, just passing a dictionary entry
    model.add(Dropout(params['dropout']))

    # again, instead of the activation name, we have a dictionary entry
    model.add(Dense(1, activation=params['last_activation']))

    # here are using a learning rate boundary
    model.compile(optimizer=params['optimizer'],
                  loss=params['losses'],
                  metrics=['acc'])

    # here we are also using the early_stopper function for a callback
    out = model.fit(x_train, y_train,
                    batch_size=params['batch_size'],
                    epochs=2,
                    verbose=0,
                    validation_data=[x_val, y_val])

    return out, model


def iris(x_train, y_train, x_val, y_val, params):

    from keras.models import Sequential
    from keras.layers import Dropout, Dense
    from talos.model import lr_normalizer, early_stopper, hidden_layers

    # note how instead of passing the value, we pass a dictionary entry
    model = Sequential()
    model.add(Dense(params['first_neuron'],
                    input_dim=x_train.shape[1],
                    activation='relu'))

    # same here, just passing a dictionary entry
    model.add(Dropout(params['dropout']))

    # with this call we can create any number of hidden layers
    hidden_layers(model, params, y_train.shape[1])

    # again, instead of the activation name, we have a dictionary entry
    model.add(Dense(y_train.shape[1],
                    activation=params['last_activation']))

    # here are using a learning rate boundary
    model.compile(optimizer=params['optimizer']
                  (lr=lr_normalizer(params['lr'],
                                    params['optimizer'])),
                  loss=params['losses'],
                  metrics=['acc'])

    # here we are also using the early_stopper function for a callback
    out = model.fit(x_train, y_train,
                    batch_size=params['batch_size'],
                    epochs=params['epochs'],
                    verbose=0,
                    validation_data=[x_val, y_val],
                    callbacks=[early_stopper(params['epochs'], mode=[1, 1])])

    return out, model



def epoch_entropy(self, history):

    '''Called from logging/logging_run.py

    Computes the entropy for epoch metric
    variation. If validation is on,
    then returns KL divergence instead of
    simple Shannon entropy. When Keras
    validation_freq is on, Shannon entropy
    is returned. Basically, all experiments
    should use validation, so Shannon is
    provided mearly as a fallback.

    '''

    import warnings
    from scipy.stats import entropy

    warnings.simplefilter('ignore')

    out = []

    # set the default entropy mode to shannon
    mode = 'shannon'

    # try to make sure each metric has validation
    if len(self._metric_keys) == len(self._val_keys):
        # make sure that the length of the arrays are same
        for i in range(len(self._metric_keys)):
            if len(history[self._metric_keys[i]]) == len(history[self._val_keys[i]]):
                mode = 'kl_divergence'
            else:
                break

    # handle the case where only shannon entropy can be used
    if mode == 'shannon':
        for i in range(len(self._metric_keys)):
            out.append(entropy(history[self._metric_keys[i]]))

    # handle the case where kl divergence can be used
    elif mode == 'kl_divergence':
        for i in range(len(self._metric_keys)):
            out.append(entropy(history[self._val_keys[i]],
                               history[self._metric_keys[i]]))

    return out



class TorchHistory:

    def __init__(self):

        self.history = {'loss': []}

    def append(self, tensor):

        '''Takes in a tensor for loss or other criterion
        from PyTorch and stores a python scalar in the history
        object. This is for unifying the `Scan()` API between
        Keras and PyTorch'''

        self.history['loss'].append(tensor.data.item())



def scan_finish(self):

    attrs_final = ['data', 'x', 'y', 'learning_entropy', 'round_times',
                   'params', 'saved_models', 'saved_weights', 'round_history']

    attrs_to_keep = attrs_final + ['random_method', 'grid_downsample',
                                   'reduction_interval', 'reduce_loss',
                                   'reduction_method', 'reduction_metric',
                                   'reduction_threshold', 'reduction_window',
                                   'experiment_name', 'round_history']

    import time
    import pandas as pd

    # create a dataframe with permutation times
    self.round_times = pd.DataFrame(self.round_times)
    self.round_times.columns = ['start', 'end', 'duration']

    # combine epoch entropy tables
    self.learning_entropy = pd.DataFrame(self.epoch_entropy)
    self.learning_entropy.columns = self._metric_keys

    # clean the results into a dataframe
    self.data = self.result

    # remove redundant columns
    keys = list(self.__dict__.keys())
    for key in keys:
        if key not in attrs_to_keep:
            delattr(self, key)

    # summarize single inputs in dictionary
    out = {}

    for key in list(self.__dict__.keys()):
        if key not in attrs_final:
            out[key] = self.__dict__[key]

    out['complete_time'] = time.strftime('%D/%H:%M')
    try:
        out['x_shape'] = self.x.shape
    # for the case when x is list
    except AttributeError:
        out['x_shape'] = 'list'

    try:
        out['y_shape'] = self.y.shape
    except AttributeError:
        out['y_shape'] = 'list'

    # final cleanup
    keys = list(self.__dict__.keys())
    for key in keys:
        if key not in attrs_final:
            delattr(self, key)

    # add details dictionary as series
    self.details = pd.Series(out)

    # add best_model

    from ..scan.scan_addon import func_best_model, func_evaluate

    self.best_model = func_best_model.__get__(self)
    self.evaluate_models = func_evaluate.__get__(self)

    # reset the index
    self.data.index = range(len(self.data))

    return self



from keras.callbacks import Callback


class ExperimentLogCallback(Callback):

    def __init__(self,
                 experiment_name,
                 params):

        '''Takes as input the name of the experiment which will be
        used for creating a .log file with the outputs and the params
        dictionary from the input model in `Scan()`

        experiment_name | str | must match the experiment_name in `Scan()`

        '''

        super(ExperimentLogCallback, self).__init__()

        import glob
        import os

        # get the experiment id first
        list_of_files = glob.glob('./' + experiment_name + '/*.csv')
        try:
            latest_file = max(list_of_files, key=os.path.getmtime)
        except ValueError:
            print("`experiment_name` has to match `Scan(experiment_name)`")

        self.name = latest_file.replace('.csv', '') + '.log'

        # rest of the config variables
        self.params = params
        self.counter = 1
        self.new_file = True

    def on_train_begin(self, logs={}):

        import random
        self.hash = hex(abs(hash(str(random.random()))))
        self.final_out = []

    def on_train_end(self, logs={}):

        f = open(self.name, 'a+')
        [f.write(','.join(map(str, i)) + '\n') for i in self.final_out]
        f.close()

    def on_epoch_begin(self, epoch, logs={}):

        self.epoch_out = []

    def on_epoch_end(self, epoch, logs={}):

        if len(self.final_out) == 0:

            try:
                open(self.name, 'r')
            except FileNotFoundError:

                self.epoch_out.append('id')
                self.epoch_out.append('epoch')

                for key in logs.keys():

                    # add to the epoch out list
                    self.epoch_out.append(key)

                self.final_out.append(self.epoch_out)
                self.epoch_out = []

        self.epoch_out.append(self.hash)
        self.epoch_out.append(epoch + 1)

        for key in logs.keys():

            # round the values
            rounded = round(logs[key], 4)

            # add to the epoch out list
            self.epoch_out.append(rounded)

        # add to the final out list
        self.final_out.append(self.epoch_out)



def mae(y_true, y_pred):
    from keras import backend as K
    return K.mean(K.abs(y_pred - y_true), axis=-1)


def mse(y_true, y_pred):
    from keras import backend as K
    return K.mean(K.square(y_pred - y_true), axis=-1)


def rmae(y_true, y_pred):
    from keras import backend as K
    return K.sqrt(K.mean(K.abs(y_pred - y_true), axis=-1))


def rmse(y_true, y_pred):
    from keras import backend as K
    return K.sqrt(K.mean(K.square(y_pred - y_true), axis=-1))


def mape(y_true, y_pred):
    from keras import backend as K
    diff = K.abs((y_true - y_pred) / K.clip(K.abs(y_true),
                                            K.epsilon(),
                                            None))
    return 100. * K.mean(diff, axis=-1)


def msle(y_true, y_pred):
    from keras import backend as K
    first_log = K.log(K.clip(y_pred, K.epsilon(), None) + 1.)
    second_log = K.log(K.clip(y_true, K.epsilon(), None) + 1.)
    return K.mean(K.square(first_log - second_log), axis=-1)


def rmsle(y_true, y_pred):
    from keras import backend as K
    first_log = K.log(K.clip(y_pred, K.epsilon(), None) + 1.)
    second_log = K.log(K.clip(y_true, K.epsilon(), None) + 1.)
    return K.sqrt(K.mean(K.square(first_log - second_log), axis=-1))


def matthews(y_true, y_pred):

    from keras import backend as K
    y_pred_pos = K.round(K.clip(y_pred, 0, 1))
    y_pred_neg = 1 - y_pred_pos

    y_pos = K.round(K.clip(y_true, 0, 1))
    y_neg = 1 - y_pos

    tp = K.sum(y_pos * y_pred_pos)
    tn = K.sum(y_neg * y_pred_neg)

    fp = K.sum(y_neg * y_pred_pos)
    fn = K.sum(y_pos * y_pred_neg)

    numerator = (tp * tn - fp * fn)
    denominator = K.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))

    return numerator / (denominator + K.epsilon())


def precision(y_true, y_pred):

    from keras import backend as K
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
    precision = true_positives / (predicted_positives + K.epsilon())
    return precision


def recall(y_true, y_pred):

    from keras import backend as K
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
    recall = true_positives / (possible_positives + K.epsilon())
    return recall


def fbeta(y_true, y_pred, beta=1):

    from keras import backend as K
    if beta < 0:
        raise ValueError('The lowest choosable beta is zero (only precision).')

    # If there are no true positives, fix the F score at 0 like sklearn.
    if K.sum(K.round(K.clip(y_true, 0, 1))) == 0:
        return 0

    p = precision(y_true, y_pred)
    r = recall(y_true, y_pred)
    bb = beta ** 2
    fbeta_score = (1 + bb) * (p * r) / (bb * p + r + K.epsilon())
    return fbeta_score


def f1score(y_true, y_pred):

    return fbeta(y_true, y_pred, beta=1)



class TalosReturnError(Exception):
    pass


class TalosParamsError(Exception):
    pass


class TalosTypeError(Exception):
    pass


class TalosModelError(Exception):
    pass


class TalosDataError(Exception):
    pass



def generator(x, y, batch_size):

    '''Creates a data generator for Keras fit_generator(). '''

    import numpy as np

    samples_per_epoch = x.shape[0]
    number_of_batches = samples_per_epoch / batch_size
    counter = 0

    while 1:

        x_batch = np.array(x[batch_size*counter:batch_size*(counter+1)]).astype('float32')
        y_batch = np.array(y[batch_size*counter:batch_size*(counter+1)]).astype('float32')
        counter += 1

        yield x_batch, y_batch

        if counter >= number_of_batches:
            counter = 0



from keras.callbacks import EarlyStopping


def early_stopper(epochs=None,
                  monitor='val_loss',
                  mode='moderate',
                  min_delta=None,
                  patience=None):

    '''EARLY STOP CALLBACK

    Helps prevent wasting time when loss is not becoming
    better. Offers two pre-determined settings 'moderate'
    and 'strict' and allows input of list with two values:

    `epochs` | int | The number of epochs for the permutation e.g. params['epochs']
    `monitor` | int | The metric to monitor for change
    `mode` | str | One of the presets `lazy`, `moderate`, `strict` or `None`
    `min_delta` | float | The limit for change at which point flag is raised
    `patience` | str | the number of epochs before termination from flag

    '''
    if mode == 'lazy':
        _es_out = EarlyStopping(monitor=monitor,
                                min_delta=0,
                                patience=int(epochs / 3),
                                verbose=0, mode='auto')

    if mode == 'moderate':
        _es_out = EarlyStopping(monitor=monitor,
                                min_delta=0,
                                patience=int(epochs / 10),
                                verbose=0, mode='auto')
    elif mode == 'strict':
        _es_out = EarlyStopping(monitor=monitor,
                                min_delta=0,
                                patience=2,
                                verbose=0, mode='auto')
    else:
        _es_out = EarlyStopping(monitor=monitor,
                                min_delta=mode[0],
                                patience=mode[1],
                                verbose=0, mode='auto')
    return _es_out



def scan_prepare(self):

    '''Includes all preparation procedures up until starting the first scan
    through scan_run()'''

    from .scan_utils import initialize_log

    self._experiment_log = initialize_log(self)

    # for the case where x_val or y_val is missing when other is present
    self.custom_val_split = False
    if (self.x_val is not None and self.y_val is None) or \
       (self.x_val is None and self.y_val is not None):
        raise RuntimeError("If x_val/y_val is inputted, other must as well.")

    elif self.x_val is not None and self.y_val is not None:
        self.custom_val_split = True

    # create reference for parameter keys
    self._param_dict_keys = sorted(list(self.params.keys()))

    # create the parameter object and move to self
    from ..parameters.ParamSpace import ParamSpace
    self.param_object = ParamSpace(params=self.params,
                                   param_keys=self._param_dict_keys,
                                   random_method=self.random_method,
                                   fraction_limit=self.fraction_limit,
                                   round_limit=self.round_limit,
                                   time_limit=self.time_limit,
                                   boolean_limit=self.boolean_limit
                                   )

    # mark that it's a first round
    self.first_round = True

    # create various stores
    self.round_history = []
    self.peak_epochs = []
    self.epoch_entropy = []
    self.round_times = []
    self.result = []
    self.saved_models = []
    self.saved_weights = []

    # handle validation split
    from ..utils.validation_split import validation_split
    self = validation_split(self)

    # set data and len
    self._data_len = len(self.x)

    return self



def local_strategy(self):

    try:
        import importlib
        importlib.reload(talos_strategy)
    except NameError:
        from talos_strategy import talos_strategy

    self = talos_strategy(self)

    return self



def initialize_log(self):

    import time
    import os

    # create the experiment folder (unless one is already there)
    try:
        path = os.getcwd()
        os.mkdir(path + '/' + self.experiment_name)
    except FileExistsError:
        pass

    self._experiment_id = time.strftime('%D%H%M%S').replace('/', '')
    _file_name = self._experiment_id + '.csv'
    _experiment_log = './' + self.experiment_name + '/' + _file_name

    f = open(_experiment_log, 'w')
    f.write('')
    f.close()

    return _experiment_log



def validation_split(self):
    """Defines the attributes `x_train`, `y_train`, `x_val` and `y_val`.
    The validation (cross-validation, aka development) sets are determined
    by the attribute val_split, which is a number in (0, 1) which determines
    the proportion of the input data to be allocated for cross-validation."""

    if self.custom_val_split:
        self.x_train = self.x
        self.y_train = self.y
        # self.x/y_val are already set

    else:
        # shuffle the data before splitting
        random_shuffle(self)

        # deduce the midway point for input data
        limit = int(len(self.x) * (1 - self.val_split))

        self.x_train = self.x[:limit]
        self.y_train = self.y[:limit]

        self.x_val = self.x[limit:]
        self.y_val = self.y[limit:]

    return self


def random_shuffle(self):

    """Randomly shuffles the datasets.
    If self.seed is set, seed the generator
    to ensure that the results are reproducible."""

    def randomize(x):

        import numpy as np

        '''Helper function to support the case
        where x consist of a list of arrays.'''

        if self.seed is not None:
            np.random.seed(self.seed)

        ix = np.arange(len(x))
        np.random.shuffle(ix)

        return ix

    if isinstance(self.x, list):

        ix = randomize(self.x[0])
        out = []

        for a in self.x:
            out.append(a[ix])
        self.x = out

    else:
        ix = randomize(self.x)
        self.x = self.x[ix]

    self.y = self.y[ix]


def kfold(x, y, folds=10, shuffled=True):

    import wrangle as wr

    if shuffled is True:
        x, y = wr.array_random_shuffle(x, y)

    out_x = []
    out_y = []

    x_len = len(x)
    step = int(x_len / folds)

    lo = 0
    hi = step

    for i in range(folds):
        out_x.append(x[lo:hi])
        out_y.append(y[lo:hi])

        lo += step
        hi += step

    return out_x, out_y



def logging_run(self, round_start, start, model_history):

    import time

    # count the duration of the round
    self._round_seconds = time.time() - start

    # set end time and log
    round_end = time.strftime('%D-%H%M%S')
    self.round_times.append([round_start, round_end, self._round_seconds])

    # handle first round only things
    if self.first_round:

        # capture the history keys for later
        self._all_keys = list(model_history.history.keys())
        self._metric_keys = [k for k in self._all_keys if 'val_' not in k]
        self._val_keys = [k for k in self._all_keys if 'val_' in k]

        # create a header column for output
        _results_header = ['round_epochs'] + self._all_keys + self._param_dict_keys
        self.result.append(_results_header)

        # save the results
        from .results import save_result
        save_result(self)

        # avoid doing this again
        self.first_round = False

    # create log and other stats
    from ..metrics.entropy import epoch_entropy
    self.epoch_entropy.append(epoch_entropy(self, model_history.history))

    # get round results to the results table and save it
    from .results import run_round_results
    _round_results = run_round_results(self, model_history)

    self.result.append(_round_results)

    from .results import save_result
    save_result(self)

    # return the Scan() self
    return self



class AutoScan:

    def __init__(self,
                 task,
                 experiment_name,
                 max_param_values=None):

        '''Configure the `AutoScan()` experiment and then use
        the property `start` in the returned class object to start
        the actual experiment.

        `task` | str | 'binary', 'multi_class', 'multi_label', or 'continuous'
        `max_param_values` | int | Number of parameter values to be included.
                                   Note, this will only work when `params` is
                                   not passed as kwargs in `AutoScan.start`.
        '''

        self.task = task
        self.max_param_values = max_param_values
        self.experiment_name = experiment_name

    def start(self, x, y, **kwargs):

        '''Start the scan. Note that you can use `Scan()` arguments as you
        would otherwise directly interacting with `Scan()`.

        `x` | array or list of arrays | prediction features
        `y` | array or list of arrays | prediction outcome variable
        `kwargs` | arguments | any `Scan()` argument can be passed here

        '''

        import talos

        m = talos.autom8.AutoModel(self.task, self.experiment_name).model

        try:
            kwargs['params']
            scan_object = talos.Scan(x, y,
                                     model=m,
                                     experiment_name=self.experiment_name,
                                     **kwargs)
        except KeyError:
            p = talos.autom8.AutoParams(task=self.task)

            if self.max_param_values is not None:
                p.resample_params(self.max_param_values)
            params = p.params
            scan_object = talos.Scan(x=x,
                                     y=y,
                                     params=params,
                                     model=m,
                                     experiment_name=self.experiment_name,
                                     **kwargs)

        return scan_object



class Restore:

    '''Restores the scan_object that had been stored locally as a result
    of talos.Deploy(scan_object, 'example')

    USE:

    diabetes = ta.Scan(x, y, p, input_model)
    ta.Deploy(diabetes, 'diabetes')
    ta.Restore('diabetes.zip')

    '''

    def __init__(self, path_to_zip):

        from zipfile import ZipFile

        import pandas as pd
        import numpy as np

        # create paths
        self.path_to_zip = path_to_zip
        self.extract_to = path_to_zip.replace('.zip', '')
        self.package_name = self.extract_to.split('/')[-1]
        self.file_prefix = self.extract_to + '/' + self.package_name

        # extract the zip
        # unpack_archive(self.path_to_zip, self.extract_to)
        z = ZipFile(self.path_to_zip, mode='r')
        z.extractall(self.extract_to)

        # add params dictionary
        self.params = np.load(self.file_prefix + '_params.npy',
                              allow_pickle=True).item()

        # add experiment details
        self.details = pd.read_csv(self.file_prefix + '_details.txt',
                                   header=None)

        # add x data sample
        self.x = pd.read_csv(self.file_prefix + '_x.csv', header=None)

        # add y data sample
        self.y = pd.read_csv(self.file_prefix + '_y.csv', header=None)

        # add model
        from talos.utils.load_model import load_model
        self.model = load_model(self.file_prefix + '_model')

        # add results
        self.results = pd.read_csv(self.file_prefix + '_results.csv')
        self.results.drop('Unnamed: 0', axis=1, inplace=True)

        # clean up
        del self.extract_to, self.file_prefix
        del self.package_name, self.path_to_zip



from keras.models import model_from_json


def load_model(saved_model):

    '''Load a Model from local disk

    Takes as input .json and .h5 file with model
    and weights and returns a model that can be then
    used for predictions.

    saved_model :: name of the saved model without
    suffix (e.g. 'iris_model' and not 'iris_model.json')

    '''

    json_file = open(saved_model + ".json", 'r')
    loaded_model_json = json_file.read()
    json_file.close()
    model = model_from_json(loaded_model_json)
    model.load_weights(saved_model + '.h5')

    return model



def run_round_results(self, out):

    '''Called from logging/logging_run.py

    THE MAIN FUNCTION FOR CREATING RESULTS FOR EACH ROUNDself.
    Takes in the history object from model.fit() and handles it.

    NOTE: The epoch level data will be dropped here each round.

    '''

    self._round_epochs = len(list(out.history.values())[0])

    _round_result_out = [self._round_epochs]

    # record the last epoch result
    for key in out.history.keys():
        _round_result_out.append(out.history[key][-1])

    # record the round hyper-parameters
    for key in self.round_params.keys():
        _round_result_out.append(self.round_params[key])

    return _round_result_out


def save_result(self):

    '''SAVES THE RESULTS/PARAMETERS TO A CSV SPECIFIC TO THE EXPERIMENT'''

    import numpy as np

    np.savetxt(self._experiment_log,
               self.result,
               fmt='%s',
               delimiter=',')


def result_todf(self):

    '''ADDS A DATAFRAME VERSION OF THE RESULTS TO THE CLASS OBJECT'''

    import pandas as pd

    # create dataframe for results
    cols = self.result[0]
    self.result = pd.DataFrame(self.result[1:])
    self.result.columns = cols

    return self


def peak_epochs_todf(self):

    import pandas as pd

    return pd.DataFrame(self.peak_epochs, columns=self.peak_epochs[0]).drop(0)



# In this init we load everything under utils in the Talos namespace

from kerasplotlib import TrainingLog as live

from ..model.normalizers import lr_normalizer
from ..model.hidden_layers import hidden_layers
from ..model.early_stopper import early_stopper
from .generator import generator
from . import gpu_utils
import talos.metrics.keras_metrics as metrics
from .sequence_generator import SequenceGenerator
from .experiment_log_callback import ExperimentLogCallback
from .torch_history import TorchHistory

del experiment_log_callback, sequence_generator



class Scan:
    """Hyperparamater scanning and optimization

    USE: ta.Scan(x=x, y=y, params=params_dict, model=model)

    Takes in a Keras model, and a dictionary with the parameter
    boundaries for the experiment.

        p = {
            'epochs' : [50, 100, 200],
            'activation' : ['relu'],
            'dropout': (0, 0.1, 5)
        }

    Accepted input formats are [1] single value in a list, [0.1, 0.2]
    multiple values in a list, and (0, 0.1, 5) a range of 5 values
    from 0 to 0.1.

    Here is an example of the input model:

    def model():

        # any Keras model

        return out, model


    You must replace the parameters in the model with references to
    the dictionary, for example:

    model.fit(epochs=params['epochs'])

    To learn more, start from the examples and documentation
    available here: https://github.com/autonomio/talos


    # CORE ARGUMENTS
    ----------------
    x : ndarray
        1d or 2d array, or a list of arrays with features for the prediction
        task.
    y : ndarray
        1d or 2d array, or a list of arrays with labels for the prediction
        task.
    params : dict
        Lists all permutations of hyperparameters, a subset of which will be
        selected at random for training and evaluation.
    model : keras model
        Any Keras model with relevant declrations like params['first_neuron']
    experiment_name : str
        Experiment name will be used to produce a folder (unless already) it's
        there from previous iterations of the experiment. Logs of the
        experiment are saved in the folder with timestamp of start
        time as filenames.
    x_val : ndarray
        User specified cross-validation data. (Default is None).
    y_val : ndarray
        User specified cross-validation labels. (Default is None).
    val_split : float, optional
        The proportion of the input `x` which is set aside as the
        validation data. (Default is 0.3).

    # RANDOMNESS ARGUMENTS
    ----------------------

    random_method : str
        Determinines the way in which the grid_downsample is applied. The
        default setting is 'uniform_mersenne'.
    seed : int
        Sets numpy random seed.

    # LIMITER ARGUMENTS
    -------------------

    performance_target : None or list [metric, threshold, loss or not]
        Allows setting a threshold for a given metric, at which point the
        experiment will be concluded as successful.
        E.g. performance_target=['f1score', 0.8, False]
    fraction_limit : int
        The fraction of `params` that will be tested (Default is None).
        Previously grid_downsample.
    round_limit : int
        Limits the number of rounds (permutations) in the experiment.
    time_limit : None or str
        Allows setting a time when experiment will be completed. Use the format
        "%Y-%m-%d %H:%M" here.
    boolean_limit : None or lambda function
        Allows setting a limit to accepted permutations as a lambda function.
        E.g. example lambda p: p['first_neuron'] * p['hidden_layers'] < 220

    # OPTIMIZER ARGUMENTS
    ---------------------
    reduction_method : None or string
        If None, random search will be used as the optimization strategy.
        Otherwise use the name of the specific strategy, e.g. 'correlation'.
    reduction_interval : None or int
        The number of reduction method rounds that will be performed. (Default
        is None).
    reduction_window : None or int
        The number of rounds of the reduction method before observing the
        results. (Default is None).
    reduction_threshold: None or float
        The minimum value for reduction to be applied. For example, when
        the 'correlation' reducer finds correlation below the threshold,
        nothing is reduced.
    reduction_metric : None or str
        Metric used to tune the reductions. minimize_loss has to be set to True
        if this is a loss.
    minimize_loss : bool
        Must be set to True if a reduction_metric is a loss.

    # OUTPUT ARGUMENTS
    ------------------
    disable_progress_bar : bool
        Disable TQDM live progress bar.
    print_params : bool
        Print params for each round on screen (useful when using TrainingLog
        callback for visualization)

    # OTHER ARGUMENTS
    -----------------
    clear_session : bool
        If the backend session is cleared between every permutation.
    save_weights : bool
        If set to False, then model weights will not be saved and best_model
        and some other features will not work. Will reduce memory pressure
        on very large models and high number of rounds/permutations.
    """

    global self

    def __init__(self,
                 x,
                 y,
                 params,
                 model,
                 experiment_name,
                 x_val=None,
                 y_val=None,
                 val_split=.3,
                 random_method='uniform_mersenne',
                 seed=None,
                 performance_target=None,
                 fraction_limit=None,
                 round_limit=None,
                 time_limit=None,
                 boolean_limit=None,
                 reduction_method=None,
                 reduction_interval=50,
                 reduction_window=20,
                 reduction_threshold=0.2,
                 reduction_metric='val_acc',
                 minimize_loss=False,
                 disable_progress_bar=False,
                 print_params=False,
                 clear_session=True,
                 save_weights=True):

        self.x = x
        self.y = y
        self.params = params
        self.model = model
        self.experiment_name = experiment_name
        self.x_val = x_val
        self.y_val = y_val
        self.val_split = val_split

        # randomness
        self.random_method = random_method
        self.seed = seed

        # limiters
        self.performance_target = performance_target
        self.fraction_limit = fraction_limit
        self.round_limit = round_limit
        self.time_limit = time_limit
        self.boolean_limit = boolean_limit

        # optimization
        self.reduction_method = reduction_method
        self.reduction_interval = reduction_interval
        self.reduction_window = reduction_window
        self.reduction_threshold = reduction_threshold
        self.reduction_metric = reduction_metric
        self.minimize_loss = minimize_loss

        # display
        self.disable_progress_bar = disable_progress_bar
        self.print_params = print_params

        # performance
        self.clear_session = clear_session
        self.save_weights = save_weights
        # input parameters section ends

        self._runtime()

    def _runtime(self):

        from .scan_run import scan_run
        self = scan_run(self)



class Analyze:

    '''A suite of commands that are useful for analyzing the results
    of a completed scan, or during a scan.

    filename :: the name of the experiment log from Scan()'''

    def __init__(self, source=None):

        '''Takes as input a filename to the experiment
        log or the Scan object'''

        import pandas as pd

        if isinstance(source, str):
            self.data = pd.read_csv(source)
        else:
            self.data = source.data

    def high(self, metric):

        '''Returns the highest value for a given metric'''

        return max(self.data[metric])

    def rounds(self):

        '''Returns the number of rounds in the experiment'''

        return len(self.data)

    def rounds2high(self, metric):

        '''Returns the number of rounds it took to get to the
        highest value for a given metric.'''

        return self.data[self.data[metric] == self.data[metric].max()].index[0]

    def low(self, metric):

        '''Returns the minimum value for a given metric'''

        return min(self.data[metric])

    def correlate(self, metric, exclude):

        '''Returns a correlation table against a given metric. Drops
        all other metrics and correlates against hyperparameters only.

        metric | str | Column label for the metric to correlate with
        exclude | list | Column label/s to be excluded from the correlation

        NOTE: You should use `exclude` to avoid correlating with other metrics.

        '''

        columns = [c for c in self.data.columns if c not in exclude + [metric]]
        out = self.data[columns]
        out.insert(0, metric, self.data[metric])

        out = out.corr()[metric]

        return out[out != 1]

    def plot_line(self, metric):

        '''A line plot for a given metric where rounds is on x-axis

        NOTE: remember to invoke %matplotlib inline if in notebook

        metric | str | Column label for the metric to correlate with

        '''
        try:
            import astetik as ast
            return ast.line(self.data, metric)
        except:
            print('Matplotlib Runtime Error. Plots will not work.')

    def plot_hist(self, metric, bins=10):

        '''A histogram for a given metric

        NOTE: remember to invoke %matplotlib inline if in notebook

        metric | str | Column label for the metric to correlate with
        bins | int | Number of bins to use in histogram

        '''
        try:
            import astetik as ast
            return ast.hist(self.data, metric, bins=bins)
        except RuntimeError:
            print('Matplotlib Runtime Error. Plots will not work.')

    def plot_corr(self, metric, exclude, color_grades=5):

        '''A heatmap with a single metric and hyperparameters.

        NOTE: remember to invoke %matplotlib inline if in notebook

        metric | str | Column label for the metric to correlate with
        exclude | list | Column label/s to be excluded from the correlation
        color_grades | int | Number of colors to use in heatmap

        '''

        try:
            import astetik as ast
            cols = self._cols(metric, exclude)
            return ast.corr(self.data[cols], color_grades=color_grades)
        except RuntimeError:
            print('Matplotlib Runtime Error. Plots will not work.')

    def plot_regs(self, x, y):

        '''A regression plot with data on two axis

        x = data for the x axis
        y = data for the y axis
        '''

        try:
            import astetik as ast
            return ast.regs(self.data, x, y)
        except RuntimeError:
            print('Matplotlib Runtime Error. Plots will not work.')

    def plot_box(self, x, y, hue=None):

        '''A box plot with data on two axis

        x = data for the x axis
        y = data for the y axis
        hue = data for the hue separation
        '''
        try:
            import astetik as ast
            return ast.box(self.data, x, y, hue)
        except RuntimeError:
            print('Matplotlib Runtime Error. Plots will not work.')

    def plot_bars(self, x, y, hue, col):

        '''A comparison plot with 4 axis'''

        try:
            import astetik as ast
            return ast.bargrid(self.data,
                               x=x,
                               y=y,
                               hue=hue,
                               col=col,
                               col_wrap=4)
        except RuntimeError:
            print('Matplotlib Runtime Error. Plots will not work.')

    def plot_kde(self, x, y=None):

        '''Kernel Destiny Estimation type histogram with
        support for 1 or 2 axis of data'''

        try:
            import astetik as ast
            return ast.kde(self.data, x, y)
        except RuntimeError:
            print('Matplotlib Runtime Error. Plots will not work.')

    def table(self, metric, exclude=[], sort_by=None, ascending=False):

        '''Shows a table with hyperparameters and a given metric

        EXAMPLE USE:

        ra1 = Reporting('diabetes_1.csv')
        ra1.table(sort_by='fmeasure_acc', ascending=False)

        PARAMS:

        metric | str or list | Column labels for the metric to correlate with
        exclude | list | Column label/s to be excluded from the correlation
        sort_by | str | The colunm name sorting should be based on
        ascending | bool | Set to True when `sort_by` is to be minimized eg. loss

        '''

        cols = self._cols(metric, exclude)

        if sort_by is None:
            sort_by = metric

        out = self.data[cols].sort_values(sort_by, ascending=ascending)

        return out

    def best_params(self, metric, exclude, n=10, ascending=False):

        '''Get the best parameters of the experiment based on a metric.
        Returns a numpy array with the values in a format that can be used
        with the talos backend in Scan(). Adds an index as the last column.

        metric | str or list | Column labels for the metric to correlate with
        exclude | list | Column label/s to be excluded from the correlation
        n | int | Number of hyperparameter permutations to be returned
        ascending | bool | Set to True when `metric` is to be minimized eg. loss

        '''

        cols = self._cols(metric, exclude)
        out = self.data[cols].sort_values(metric, ascending=ascending)
        out = out.drop(metric, axis=1).head(n)
        out.insert(out.shape[1], 'index_num', range(len(out)))

        return out.values

    def _cols(self, metric, exclude):

        '''Helper to remove other than desired metric from data table'''

        cols = [col for col in self.data.columns if col not in exclude + [metric]]

        if isinstance(metric, list) is False:
            metric = [metric]
        for i, metric in enumerate(metric):
            cols.insert(i, metric)

        # make sure only unique values in col list
        cols = list(set(cols))

        return cols



def hidden_layers(model, params, last_neuron):

    '''HIDDEN LAYER Generator

    NOTE: 'first_neuron', 'dropout', and 'hidden_layers' need
    to be present in the params dictionary.

    Hidden layer generation for the cases where number
    of layers is used as a variable in the optimization process.
    Handles things in a way where any number of layers can be tried
    with matching hyperparameters.'''

    from keras.layers import Dense, Dropout
    from .network_shape import network_shape
    from ..utils.exceptions import TalosParamsError

    try:
        kernel_initializer = params['kernel_initializer']
    except KeyError:
        kernel_initializer = 'glorot_uniform'

    try:
        kernel_regularizer = params['kernel_regularizer']
    except KeyError:
        kernel_regularizer = None

    try:
        bias_initializer = params['bias_initializer']
    except KeyError:
        bias_initializer = 'zeros'

    try:
        bias_regularizer = params['bias_regularizer']
    except KeyError:
        bias_regularizer = None

    try:
        use_bias = params['use_bias']
    except KeyError:
        use_bias = True

    try:
        activity_regularizer = params['activity_regularizer']
    except KeyError:
        activity_regularizer = None

    try:
        kernel_constraint = params['kernel_constraint']
    except KeyError:
        kernel_constraint = None

    try:
        bias_constraint = params['bias_constraint']
    except KeyError:
        bias_constraint = None

    # check for the params that are required for hidden_layers
    for param in ['shapes', 'first_neuron', 'dropout']:
        try:
            params[param]
        except KeyError as err:
            if err.args[0] == param:
                raise TalosParamsError("hidden_layers requires '" + param + "' in params")

    layer_neurons = network_shape(params, last_neuron)

    for i in range(params['hidden_layers']):

        model.add(Dense(layer_neurons[i],
                        activation=params['activation'],
                        use_bias=use_bias,
                        kernel_initializer=kernel_initializer,
                        kernel_regularizer=kernel_regularizer,
                        bias_initializer=bias_initializer,
                        bias_regularizer=bias_regularizer,
                        activity_regularizer=activity_regularizer,
                        kernel_constraint=kernel_constraint,
                        bias_constraint=bias_constraint))

        model.add(Dropout(params['dropout']))



def output_layer(task, last_activation, y_train, y_val):

    import numpy as np

    # output layer
    if task == 'binary':
        activation = last_activation
        last_neuron = 1

    elif task == 'multi_class':
        activation = last_activation
        last_neuron = len(np.unique(np.hstack((y_train, y_val))))

    elif task == 'multi_label':
        activation = last_activation
        last_neuron = y_train.shape[1]

    elif task == 'continuous':
        activation = None
        last_neuron = 1

    elif isinstance(task, list):
        try:
            # multilabel
            activation = last_activation
            last_neuron = y_train.shape[1]
        except IndexError:
            uniques = np.unique(np.hstack((y_train, y_val)))
            # binary
            if uniques == 2:
                activation = last_activation
                last_neuron = 1
            # multiclass (note this supports only < 10 classes)
            elif uniques <= 10:
                activation = last_activation
                last_neuron = np.unique(np.hstack((y_train, y_val)))
            # continuous maybe, or too many classes
            else:
                activation = None
                last_neuron = 1

    return activation, last_neuron



class Predict:

    '''Class for making predictions on the models that are stored
    in the Scan() object'''

    def __init__(self, scan_object, task):

        '''Takes in as input a Scan() object'''

        self.scan_object = scan_object
        self.data = scan_object.data
        self.task = task

    def predict(self, x, model_id=None, metric='val_acc', asc=False):

        '''Makes a probability prediction from input x. If model_id
        is not given, then best_model will be used.'''

        if model_id is None:
            from ..utils.best_model import best_model
            model_id = best_model(self.scan_object, metric, asc)

        from ..utils.best_model import activate_model
        model = activate_model(self.scan_object, model_id)

        return model.predict(x)

    def predict_classes(self, x, model_id=None, metric='val_acc', asc=False):

        '''Makes a class prediction from input x. If model_id
        is not given, then best_model will be used.'''

        if model_id is None:
            from ..utils.best_model import best_model
            model_id = best_model(self.scan_object, metric, asc)

        from ..utils.best_model import activate_model
        model = activate_model(self.scan_object, model_id)

        return model.predict_classes(x)



from keras.optimizers import SGD, Adam, Adadelta, Adagrad, Adamax, RMSprop
from keras.optimizers import Nadam


def lr_normalizer(lr, optimizer):
    """Assuming a default learning rate 1, rescales the learning rate
    such that learning rates amongst different optimizers are more or less
    equivalent.

    Parameters
    ----------
    lr : float
        The learning rate.
    optimizer : keras optimizer
        The optimizer. For example, Adagrad, Adam, RMSprop.
    """

    if optimizer == Adadelta:
        pass
    elif optimizer == SGD or optimizer == Adagrad:
        lr /= 100.0
    elif optimizer == Adam or optimizer == RMSprop:
        lr /= 1000.0
    elif optimizer == Adamax or optimizer == Nadam:
        lr /= 500.0

    return lr



class Evaluate:

    '''Class for evaluating models based on the Scan() object'''

    def __init__(self, scan_object):

        '''Takes in as input a Scan() object.
        e = evaluate(scan_object) and see docstring
        for e() for more information.'''

        self.scan_object = scan_object
        self.data = scan_object.data

    def evaluate(self,
                 x,
                 y,
                 task,
                 metric,
                 model_id=None,
                 folds=5,
                 shuffle=True,
                 asc=False,
                 print_out=False):

        '''Evaluate a model based on f1_score (all except regression)
        or mae (for regression). Supports 'binary', 'multi_class',
        'multi_label', and 'regression' evaluation.

        x : array
            The input data for making predictions
        y : array
            The ground truth for x
        model_id : int
            It's possible to evaluate a specific model based on ID.
            Can be None.
        folds : int
            Number of folds to use for cross-validation
        sort_metric : string
            A column name referring to the metric that was used in the
            scan_object as a performance metric. This is used for sorting
            the results to pick for evaluation.
        shuffle : bool
            Data is shuffled before evaluation.
        task : string
            'binary', 'multi_class', 'multi_label', or 'continuous'.
        asc : bool
            False if the metric is to be optimized upwards
            (e.g. accuracy or f1_score)
        print_out : bool
            Print out the results.

        TODO: add possibility to input custom metrics.

        '''

        import numpy as np
        import sklearn as sk

        out = []
        if model_id is None:
            from ..utils.best_model import best_model
            model_id = best_model(self.scan_object, metric, asc)

        from ..utils.best_model import activate_model
        model = activate_model(self.scan_object, model_id)

        from ..utils.validation_split import kfold
        kx, ky = kfold(x, y, folds, shuffle)

        for i in range(folds):

            y_pred = model.predict(kx[i], verbose=0)

            if task == 'binary':
                y_pred = y_pred >= .5
                scores = sk.metrics.f1_score(y_pred, ky[i], average='binary')

            elif task == 'multi_class':
                y_pred = y_pred.argmax(axis=-1)
                print(y_pred)
                print(ky[i])
                scores = sk.metrics.f1_score(y_pred, ky[i], average='macro')

            if task == 'multi_label':
                y_pred = model.predict(kx[i]).argmax(axis=1)
                scores = sk.metrics.f1_score(y_pred,
                                             ky[i].argmax(axis=1),
                                             average='macro')

            elif task == 'continuous':
                y_pred = model.predict(kx[i])
                scores = sk.metrics.mean_absolute_error(y_pred, ky[i])

            out.append(scores)

        if print_out is True:
            print("mean : %.2f \n std : %.2f" % (np.mean(out), np.std(out)))

        return out



def telco_churn(quantile=.5):

    '''Returns dataset in format x, [y1, y2]. This dataset
    is useful for demonstrating multi-output model or for
    experimenting with reduction strategy creation.

    The data is from hyperparameter optimization experiment with
    Kaggle telco churn dataset.

    x: features
    y1: val_loss
    y2: val_f1score

    quantile is for transforming the otherwise continuous y variables into
    labels so that higher value is stronger. If set to 0 then original
    continuous will be returned.'''

    import wrangle
    import pandas as pd

    df = pd.read_csv('https://raw.githubusercontent.com/autonomio/examples/master/telco_churn/telco_churn_for_sensitivity.csv')

    df = df.drop(['val_acc', 'loss', 'f1score', 'acc', 'round_epochs'], 1)

    for col in df.iloc[:, 2:].columns:
        df = wrangle.col_to_multilabel(df, col)

    df = wrangle.df_rename_cols(df)

    if quantile > 0:
        y1 = (df.C0 < df.C0.quantile(quantile)).astype(int).values
        y2 = (df.C1 > df.C1.quantile(quantile)).astype(int).values
    else:
        y1 = df.C0.values
        y2 = df.C1.values

    x = df.drop(['C0', 'C1'], 1).values

    return x, [y1, y2]


def icu_mortality(samples=None):

    import pandas as pd
    base = 'https://raw.githubusercontent.com/autonomio/datasets/master/autonomio-datasets/'
    df = pd.read_csv(base + 'icu_mortality.csv')
    df = df.dropna(thresh=3580, axis=1)
    df = df.dropna()
    df = df.sample(frac=1).head(samples)
    y = df['hospitalmortality'].astype(int).values
    x = df.drop('hospitalmortality', axis=1).values

    return x, y


def titanic():

    import pandas as pd
    base = 'https://raw.githubusercontent.com/autonomio/datasets/master/autonomio-datasets/'
    df = pd.read_csv(base + 'titanic.csv')

    y = df.survived.values

    x = df[['age', 'sibsp', 'parch']]
    cols = ['class', 'embark_town', 'who', 'deck', 'sex']

    for col in cols:
        x = pd.merge(x,
                     pd.get_dummies(df[col]),
                     left_index=True,
                     right_index=True)

    x = x.values

    print('BE CAREFUL, this dataset has nan values.')

    return x, y


def iris():

    import pandas as pd
    from keras.utils import to_categorical
    base = 'https://raw.githubusercontent.com/autonomio/datasets/master/autonomio-datasets/'
    df = pd.read_csv(base + 'iris.csv')
    df['species'] = df['species'].factorize()[0]
    df = df.sample(len(df))
    y = to_categorical(df['species'])
    x = df.iloc[:, :-1].values

    y = to_categorical(df['species'])
    x = df.iloc[:, :-1].values

    return x, y


def cervical_cancer():

    import pandas as pd
    from numpy import nan
    base = 'https://raw.githubusercontent.com/autonomio/datasets/master/autonomio-datasets/'
    df = pd.read_csv(base + 'cervical_cancer.csv')
    df = df.replace('?', nan)
    df = df.drop(['citology', 'hinselmann', 'biopsy'], axis=1)
    df = df.drop(['since_first_diagnosis',
                  'since_last_diagnosis'], axis=1).dropna()

    df = df.astype(float)

    y = df.schiller.values
    x = df.drop('schiller', axis=1).values

    return x, y


def breast_cancer():

    import pandas as pd
    base = 'https://raw.githubusercontent.com/autonomio/datasets/master/autonomio-datasets/'
    df = pd.read_csv(base + 'breast_cancer.csv')

    # then some minimal data cleanup
    df.drop("Unnamed: 32", axis=1, inplace=True)
    df.drop("id", axis=1, inplace=True)

    # separate to x and y
    y = df.diagnosis.values
    x = df.drop('diagnosis', axis=1).values

    # convert the string labels to binary
    y = (y == 'M').astype(int)

    return x, y


def mnist():

    '''Note that this dataset, unlike other Talos datasets,returns:

    x_train, y_train, x_val, y_val'''

    import keras
    import numpy as np

    # the data, split between train and test sets
    (x_train, y_train), (x_val, y_val) = keras.datasets.mnist.load_data()

    # input image dimensions
    img_rows, img_cols = 28, 28

    if keras.backend.image_data_format() == 'channels_first':

        x_train = x_train.reshape(x_train.shape[0], 1, img_rows, img_cols)
        x_val = x_val.reshape(x_val.shape[0], 1, img_rows, img_cols)
        input_shape = (1, img_rows, img_cols)

    else:
        x_train = x_train.reshape(x_train.shape[0], img_rows, img_cols, 1)
        x_val = x_val.reshape(x_val.shape[0], img_rows, img_cols, 1)
        input_shape = (img_rows, img_cols, 1)

    x_train = x_train.astype('float32')
    x_val = x_val.astype('float32')
    x_train /= 255
    x_val /= 255

    classes = len(np.unique(y_train))

    # convert class vectors to binary class matrices
    y_train = keras.utils.to_categorical(y_train, classes)
    y_val = keras.utils.to_categorical(y_val, classes)

    print("Use input_shape %s" % str(input_shape))

    return x_train, y_train, x_val, y_val



class GamifyMap:

    def __init__(self, scan_object):

        '''GamifyMap handles the management of the
        dictionary that contains the information about
        hyperparameters, which is exchanged in and out
        during the `Scan()` experiment.
        '''

        self.params = scan_object.param_object.params
        self.scan_object = scan_object
        self.generate_gamify_dict()
        self.gamify_map = self.generate_gamify_dict_map()

        # parse together the output file name
        _folder = './' + scan_object.experiment_name + '/'
        _id = scan_object._experiment_id
        self._filename = _folder + _id

    def run_updates(self):

        for key in self.gamify_dict.keys():
            for val in self.gamify_dict[key].keys():
                if self.gamify_dict[key][val] != self.updated_dict[key][val]:

                    label = list(self.params.keys())[int(key)]
                    value = self.params[label][int(val)]

                    self.gamify_dict[key][val] = self.updated_dict[key][val]
                    self.scan_object.param_object.remove_is(label, value)

        return self.scan_object

    def back_to_original(self, gamify_from_json):

        gamify_dict = {}
        for key in gamify_from_json:
            param_vals = {}
            for val in gamify_from_json[key]:
                param_vals[g.gamify_map[key][val][1]] = gamify_from_json[key][val][:2]
            gamify_dict[g.gamify_map[key][val][0]] = param_vals

        return gamify_dict

    def generate_gamify_dict_map(self):

        gamify_dict_map = {}

        for i, key in enumerate(self.params.keys()):
            param_vals = {}
            for ii, val in enumerate(self.params[key]):
                param_vals[str(ii)] = [key, val]
            gamify_dict_map[str(i)] = param_vals

        return gamify_dict_map

    def generate_gamify_dict(self):

        '''This is done once at the beginning
        of the experiment.

        NOTE: This will be all stringified, so an index
        mapping system will be used to convert back to
        actual forms later.'''

        gamify_dict = {}

        for i, key in enumerate(self.params.keys()):
            param_vals = {}
            for ii, val in enumerate(self.params[key]):
                param_vals[str(ii)] = ['active', 0, str(key), str(val)]
            gamify_dict[str(i)] = param_vals

        self.gamify_dict = gamify_dict

    def export_json(self):

        import json

        with open(self._filename + '.json', 'w') as fp:
            json.dump(self.gamify_dict, fp)

    def import_json(self):

        import json

        with open(self._filename + '.json', 'r') as fp:
            out = json.load(fp)

        self.updated_dict = out



def AutoPredict(scan_object,
                x_val,
                y_val,
                x_pred,
                task,
                metric='val_acc',
                n_models=10,
                folds=5,
                shuffle=True,
                asc=False):

    '''Automatically handles the process of finding the best models from a
    completed `Scan()` experiment, evaluates those models, and uses the winner
    to make predictions on input data.

    NOTE: the input data must be in same format as 'x' that was
    used in `Scan()`.

    Parameters
    ----------
    scan_object : Scan() object
        A Scan() process needs to be completed first, and then the resulting
        object can be used as input here.
    x_val : ndarray or list of ndarray
        Data to be used for 'x' in evaluation. Note that should be in the same
        format as the data which was used in the Scan() but not the same data.
    y_val : ndarray or list of ndarray
        Data to be used for 'y' in evaluation. Note that should be in the same
        format as the data which was used in the Scan() but not the same data.
    y_pred : ndarray or list of ndarray
        Input data to be used for the actual predictions in evaluation. Note
        it should be in the same format as the data which was used in the
        Scan() but not the same data.
    task : string
        'binary', 'multi_class', 'multi_label', or 'continuous'.
    metric : str
        The metric to be used for deciding which models are promising.
        Basically the 'n' argument and 'metric' argument are combined to pick
        'n' best performing models based on 'metric'.
    n_models : str
        Number of promising models to be included in the evaluation process.
        Time increase linearly with number of models.
    folds : int
        Number of folds to be used in cross-validation.
    shuffle : bool
        If the data should be shuffled before cross-validation.
    average : str
        This parameter is required for multiclass/multilabel targets. If None,
        the scores for each class are returned. Otherwise, this determines
        the type of averaging performed on the data:

        'binary':
        Only report results for the class specified by pos_label.
        This is applicable only if targets (y_{true,pred}) are binary.

        'micro':
        Calculate metrics globally by counting the total true positives,
        false negatives and false positives.

        'macro':
        Calculate metrics for each label, and find their unweighted mean.
        This does not take label imbalance into account.

        'weighted':
        Calculate metrics for each label, and find their average weighted
        by support (the number of true instances for each label). This alters
        'macro' to account for label imbalance; it can result in an F-score
        that is not between precision and recall.

        'samples':
        Calculate metrics for each instance, and find their average
        (only meaningful for multilabel classification where this differs
        from accuracy_score).
    asc : bool
        This needs to be True for evaluation metrics that need to be minimized,
        and False when a metric needs to be maximized.

    '''

    # evaluate and add the evaluation scores
    scan_object.evaluate_models(x_val,
                                y_val,
                                n_models=n_models,
                                task=task,
                                metric=metric,
                                folds=folds,
                                shuffle=shuffle,
                                asc=False)

    # get the best model based on evaluated score
    scan_object.preds_model = scan_object.best_model('eval_f1score_mean')

    # make predictions with the model
    scan_object.preds_probabilities = scan_object.preds_model.predict(x_pred)

    # make (class) predictiosn with the model
    scan_object.preds_classes = scan_object.preds_model.predict_classes(x_pred)

    # get the hyperparameter for the model
    scan_object.preds_parameters = scan_object.data.sort_values('eval_f1score_mean',
                                                           ascending=False).iloc[0]

    print(">> Added model, probabilities, classes, and parameters to scan_object")

    return scan_object



def gamify(self):

    '''Will apply reduction changes based on edits on the
    the produced .json file in the experiment folder'''

    if self.param_object.round_counter == 1:

        # create the gamify object
        from .GamifyMap import GamifyMap
        g = GamifyMap(self)

        # keep in scan_object
        self._gamify_object = g

        # do the first export in the experiment folder
        g.export_json()

        return self

    # for every round check if there are changes
    self._gamify_object.import_json()
    self = self._gamify_object.run_updates()
    self._gamify_object.export_json()

    return self



def correlation(self, method):

    '''This is called from reduce_run.py.

    Performs a spearman rank order correlation
    based reduction. First looks for a parameter
    that correlates with reduction_metric and
    correlation meets reduction_threshold and
    then converts the match parameter into
    a 2d multilabel shape. Then new correlation
    against reduction_metric is performed to identify
    which particular value is to be dropped.

    '''

    import numpy as np

    # transform the data properly first
    from .reduce_utils import cols_to_multilabel
    data = cols_to_multilabel(self)

    # get the correlations
    corr_values = data.corr(method)[self.reduction_metric]

    # drop the reduction metric row
    corr_values.drop(self.reduction_metric, inplace=True)

    # drop labels where value is NaN
    corr_values.dropna(inplace=True)

    # if all nans, then stop
    if len(corr_values) <= 1:
        return self

    # sort based on the metric type
    corr_values.sort_values(ascending=self.minimize_loss, inplace=True)

    # if less than threshold, then stop
    if abs(corr_values[-1]) < self.reduction_threshold:
        return self

    # get the strongest correlation
    corr_values = corr_values.index[-1]

    # get the label, value, and dtype from the column header
    label, dtype, value = corr_values.split('~')

    # convert things back to their original dtype
    value = np.array([value]).astype(dtype)[0]

    # this is where we modify the parameter space accordingly
    self.param_object.remove_is(label, value)

    return self



class Deploy:

    '''Functionality for deploying a model to a filename'''

    def __init__(self, scan_object, model_name, metric, asc=False):

        '''Deploy a model to be used later or in a different system.

        NOTE: for a metric that is to be minimized, set asc=True or otherwise
        you will end up with the model that has the highest loss.

        Deploy() takes in the object from Scan() and creates a package locally
        that can be later activated with Restore().

        scan_object : object
            The object that is returned from Scan() upon completion.
        model_name : str
            Name for the .zip file to be created.
        metric : str
            The metric to be used for picking the best model.
        asc: bool
            Make this True for metrics that are to be minimized (e.g. loss) ,
            and False when the metric is to be maximized (e.g. acc)

        '''

        import os

        self.scan_object = scan_object
        os.mkdir(model_name)
        self.path = model_name + '/' + model_name
        self.model_name = model_name
        self.metric = metric
        self.asc = asc
        self.data = scan_object.data

        from ..utils.best_model import best_model, activate_model
        self.best_model = best_model(scan_object, metric, asc)
        self.model = activate_model(scan_object, self.best_model)

        # runtime
        self.save_model_as()
        self.save_details()
        self.save_data()
        self.save_results()
        self.save_params()
        self.save_readme()
        self.package()

    def save_model_as(self):

        '''Model Saver
        WHAT: Saves a trained model so it can be loaded later
        for predictions by predictor().
        '''

        model_json = self.model.to_json()
        with open(self.path + "_model.json", "w") as json_file:
            json_file.write(model_json)

        self.model.save_weights(self.path + "_model.h5")
        print("Deploy package" + " " + self.model_name + " " + "have been saved.")

    def save_details(self):

        self.scan_object.details.to_csv(self.path + '_details.txt')

    def save_data(self):

        import pandas as pd

        # input data is <= 2d
        try:
            x = pd.DataFrame(self.scan_object.x[:100])
            y = pd.DataFrame(self.scan_object.y[:100])

        # input data is > 2d
        except ValueError:
            x = pd.DataFrame()
            y = pd.DataFrame()
            print("data is not 2d, dummy data written instead.")

        x.to_csv(self.path + '_x.csv', header=None, index=None)
        y.to_csv(self.path + '_y.csv', header=None, index=None)

    def save_results(self):

        self.scan_object.data.to_csv(self.path + '_results.csv')

    def save_params(self):

        import numpy as np

        np.save(self.path + '_params', self.scan_object.params)

    def save_readme(self):

        txt = 'To activate the assets in the Talos deploy package: \n\n   from talos.commands.restore import Restore \n   a = Restore(\'path_to_asset\')\n\nNow you will have an object similar to the Scan object, which can be used with other Talos commands as you would be able to with the Scan object'

        text_file = open(self.path.split('/')[0] + '/README.txt', "w")
        text_file.write(txt)
        text_file.close()

    def package(self):

        import shutil

        shutil.make_archive(self.model_name, 'zip', self.model_name)
        shutil.rmtree(self.model_name)



from keras.models import model_from_json


def best_model(self, metric, asc):

    '''Picks the best model based on a given metric and
    returns the index number for the model.

    NOTE: for loss 'asc' should be True'''

    best = self.data.sort_values(metric, ascending=asc).iloc[0].name

    return best


def activate_model(self, model_id):

    '''Loads the model from the json that is stored in the Scan object'''

    model = model_from_json(self.saved_models[model_id])
    model.set_weights(self.saved_weights[model_id])

    return model



def trees(self, quantile=.8):

    '''Extra Trees based reduction strategy. Like 'forrest', somewhat more
    aggressive than for example 'spearman' because there are no
    negative values, but instead the highest positive correlation
    is minused from all the values so that max value is 0, and then
    values are turned into positive. The one with the highest positive
    score in the end will be dropped. This means that anything with
    0 originally, is a candidate for dropping. Because there are multiple
    zeroes in many cases, there is an element of randomness on which one
    is dropped.

    '''

    import wrangle
    import numpy as np

    # handle conversion to multi_labels
    from .reduce_utils import cols_to_multilabel
    data = cols_to_multilabel(self)

    # because extra trees wants label as 'y' we first transform with quantile
    quantile_value = data[self.reduction_metric].quantile(quantile)
    data[self.reduction_metric] = data[self.reduction_metric] > quantile_value

    # get the correlations
    corr_values = wrangle.df_corr_extratrees(data, self.reduction_metric)

    # drop labels where value is NaN
    corr_values.dropna(inplace=True)

    # handle the turning around of values (see docstring for more info)
    corr_values -= corr_values[0]
    corr_values = corr_values.abs()

    # get the strongest correlation
    corr_values = corr_values.index[-1]

    # get the label, value, and dtype from the column header
    label, dtype, value = corr_values.split('~')

    # convert things back to their original dtype
    value = np.array([value]).astype(dtype)[0]

    # this is where we modify the parameter space accordingly
    self.param_object.remove_is(label, value)

    return value, label



class AutoModel:

    def __init__(self, task, experiment_name, metric=None):

        '''

        Creates an input model for Scan(). Optimized for being used together
        with Params(). For example:

        p = talos.AutoParams().params
        model = talos.AutoModel(task='binary').model

        talos.Scan(x, y, p, model)

        NOTE: the parameter space from Params() is very large, so use limits
        in or reducers in Scan() accordingly.

        task : string or None
            If 'continuous' then mae is used for metric, if 'binary',
            'multiclass', or 'multilabel', f1score is used. Accuracy is always
            used.
        experiment_name | str | Must be same as in `Scan()`
        metric : None or list
            You can also input a list with one or more custom metrics or names
            of Keras or Talos metrics.
        '''

        from talos.utils.experiment_log_callback import ExperimentLogCallback

        self.task = task
        self.experiment_name = experiment_name
        self.metric = metric

        if self.task is not None:
            self.metrics = self._set_metric()
        elif self.metric is not None and isinstance(self.metric, list):
            self.metrics = self.metric + ['acc']
        else:
            print("Either pick task or provide list as input for metric.")

        # create the model
        self.model = self._create_input_model
        self.callback = ExperimentLogCallback

    def _set_metric(self):

        """Sets the metric for the model based on the experiment type
        or a list of metrics from user."""

        import talos as ta

        if self.task in ['binary', 'multiclass', 'multilabel']:
            return [ta.utils.metrics.f1score, 'acc']
        elif self.task == 'continuous':
            return [ta.utils.metrics.mae, 'acc']

    def _create_input_model(self, x_train, y_train, x_val, y_val, params):

        import wrangle as wr

        from keras.models import Sequential
        from keras.layers import Dropout, Flatten
        from keras.layers import LSTM, Conv1D, SimpleRNN, Dense, Bidirectional

        model = Sequential()

        if params['network'] != 'dense':
            x_train = wr.array_reshape_conv1d(x_train)
            x_val = wr.array_reshape_conv1d(x_val)

        if params['network'] == 'conv1d':
            model.add(Conv1D(params['first_neuron'], x_train.shape[1]))
            model.add(Flatten())

        elif params['network'] == 'lstm':
            model.add(LSTM(params['first_neuron']))

        if params['network'] == 'bidirectional_lstm':
            model.add(Bidirectional(LSTM(params['first_neuron'])))

        elif params['network'] == 'simplernn':
            model.add(SimpleRNN(params['first_neuron']))

        elif params['network'] == 'dense':
            model.add(Dense(params['first_neuron'],
                            input_dim=x_train.shape[1],
                            activation='relu',
                            kernel_initializer=params['kernel_initializer']))

        model.add(Dropout(params['dropout']))

        # add hidden layers to the model
        from talos.model.hidden_layers import hidden_layers
        hidden_layers(model, params, 1)

        # get the right activation and last_neuron based on task
        from talos.model.output_layer import output_layer
        activation, last_neuron = output_layer(self.task,
                                               params['last_activation'],
                                               y_train,
                                               y_val)

        model.add(Dense(last_neuron,
                        activation=activation,
                        kernel_initializer=params['kernel_initializer']))

        # bundle the optimizer with learning rate changes
        from talos.model.normalizers import lr_normalizer
        optimizer = params['optimizer'](lr=lr_normalizer(params['lr'],
                                                         params['optimizer']))

        # compile the model
        model.compile(optimizer=optimizer,
                      loss=params['losses'],
                      metrics=self.metrics)

        # fit the model
        out = model.fit(x_train, y_train,
                        batch_size=params['batch_size'],
                        epochs=params['epochs'],
                        verbose=0,
                        callbacks=[self.callback(self.experiment_name, params)],
                        validation_data=[x_val, y_val])

        # pass the output to Talos
        return out, model



from keras.utils import Sequence


class SequenceGenerator(Sequence):

    def __init__(self, x_set, y_set, batch_size):

        self.x, self.y = x_set, y_set
        self.batch_size = batch_size

    def __len__(self):

        import numpy as np

        return int(np.ceil(len(self.x) / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_x = self.x[idx * self.batch_size:(idx + 1) * self.batch_size]
        batch_y = self.y[idx * self.batch_size:(idx + 1) * self.batch_size]

        return batch_x, batch_y



def sample_reducer(limit, max_value, random_method):

    '''Sample Reducer (Helper)

    NOTE: The Scan() object  is in self.main_self because
    the object being passed here is ParamGrid() object where
    the Scan() object is attached as self.main_self.

    Utilize 'grid_downsample', 'shuffle', and 'random_method'
    to reduce the param_grid before starting the experiment.
    This is the simplest method in Talos for dealing with curse
    of dimensionality.

    Options are uniform random, stratified random, latin hypercube
    sampling, and latin hypercube with sudoku style constraint.

    Returns the reduced param_grid as numpy array.

    '''

    import chances as ch

    # calculate the size of the downsample
    if isinstance(limit, float):
        n = int(max_value * limit)
    if isinstance(limit, int):
        n = limit

    # throw an error if
    from ..utils.exceptions import TalosDataError
    if n < 1:
        raise TalosDataError("Limiters lead to < 1 permutations.")

    # Initialize Randomizer()
    r = ch.Randomizer(max_value, n)

    # use the user selected method
    if random_method == 'sobol':
        out = r.sobol()
    elif random_method == 'quantum':
        out = r.quantum()
    elif random_method == 'halton':
        out = r.halton()
    elif random_method == 'korobov_matrix':
        out = r.korobov_matrix()
    elif random_method == 'latin_sudoku':
        out = r.latin_sudoku()
    elif random_method == 'latin_matrix':
        out = r.latin_matrix()
    elif random_method == 'latin_improved':
        out = r.latin_improved()
    elif random_method == 'uniform_mersenne':
        out = r.uniform_mersenne()
    elif random_method == 'uniform_crypto':
        out = r.uniform_crypto()
    elif random_method == 'ambience':
        out = r.ambience()
    else:
        print('No eligble random_method found. Using uniform_mersenne.')
        out = r.uniform_mersenne()

    return out



def breast_cancer(round_limit=2, random_method='uniform_mersenne'):

    '''Performs a Scan with Iris dataset and simple dense net'''
    import talos as ta
    scan_object = ta.Scan(ta.templates.datasets.breast_cancer()[0],
                          ta.templates.datasets.breast_cancer()[1],
                          ta.templates.params.breast_cancer(),
                          ta.templates.models.breast_cancer,
                          'test',
                          round_limit=round_limit)

    return scan_object


def cervical_cancer(round_limit=2, random_method='uniform_mersenne'):

    '''Performs a Scan with Iris dataset and simple dense net'''
    import talos as ta
    scan_object = ta.Scan(ta.templates.datasets.cervical_cancer()[0],
                          ta.templates.datasets.cervical_cancer()[1],
                          ta.templates.params.cervical_cancer(),
                          ta.templates.models.cervical_cancer,
                          'test',
                          round_limit=round_limit)

    return scan_object


def iris(round_limit=2, random_method='uniform_mersenne'):

    '''Performs a Scan with Iris dataset and simple dense net'''
    import talos as ta
    scan_object = ta.Scan(ta.templates.datasets.iris()[0],
                          ta.templates.datasets.iris()[1],
                          ta.templates.params.iris(),
                          ta.templates.models.iris,
                          'test',
                          round_limit=round_limit)

    return scan_object


def titanic(round_limit=2, random_method='uniform_mersenne'):

    '''Performs a Scan with Iris dataset and simple dense net'''
    import talos as ta
    scan_object = ta.Scan(ta.templates.datasets.titanic()[0][:50],
                          ta.templates.datasets.titanic()[1][:50],
                          ta.templates.params.titanic(),
                          ta.templates.models.titanic,
                          'test',
                          round_limit=round_limit)

    return scan_object



#! /usr/bin/env python
#
# Copyright (C) 2018 Mikko Kotila

DESCRIPTION = "Talos Hyperparameter Tuning for Keras"
LONG_DESCRIPTION = """\
Talos radically changes the ordinary Keras workflow by
fully automating hyperparameter tuning and model evaluation.
Talos exposes Keras functionality entirely and there is no new
syntax or templates to learn.
"""

DISTNAME = 'talos'
MAINTAINER = 'Mikko Kotila'
MAINTAINER_EMAIL = 'mailme@mikkokotila.com'
URL = 'http://autonom.io'
LICENSE = 'MIT'
DOWNLOAD_URL = 'https://github.com/autonomio/talos/'
VERSION = '0.6.3'

try:
    from setuptools import setup
    _has_setuptools = True
except ImportError:
    from distutils.core import setup

install_requires = ['wrangle',
                    'numpy',
                    'pandas',
                    'keras',
                    'astetik',
                    'sklearn',
                    'tqdm',
                    'chances',
                    'kerasplotlib',
                    'requests']


if __name__ == "__main__":

    setup(name=DISTNAME,
          author=MAINTAINER,
          author_email=MAINTAINER_EMAIL,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          long_description=LONG_DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=VERSION,
          download_url=DOWNLOAD_URL,
          install_requires=install_requires,
          packages=['talos',
                    'talos.scan',
                    'talos.templates',
                    'talos.utils',
                    'talos.model',
                    'talos.parameters',
                    'talos.reducers',
                    'talos.metrics',
                    'talos.commands',
                    'talos.logging',
                    'talos.autom8'],

          classifiers=['Intended Audience :: Science/Research',
                       'Programming Language :: Python :: 2.7',
                       'Programming Language :: Python :: 3.5',
                       'Programming Language :: Python :: 3.6',
                       'License :: OSI Approved :: MIT License',
                       'Topic :: Scientific/Engineering :: Human Machine Interfaces',
                       'Topic :: Scientific/Engineering :: Artificial Intelligence',
                       'Topic :: Scientific/Engineering :: Mathematics',
                       'Operating System :: POSIX',
                       'Operating System :: Unix',
                       'Operating System :: MacOS',
                       'Operating System :: Microsoft :: Windows :: Windows 10'])



import inspect

import numpy as np
import itertools as it
from datetime import datetime


class ParamSpace:

    def __init__(self,
                 params,
                 param_keys,
                 random_method='uniform_mersenne',
                 fraction_limit=None,
                 round_limit=None,
                 time_limit=None,
                 boolean_limit=None):

        # set all the arguments
        self.params = params
        self.param_keys = param_keys
        self.fraction_limit = fraction_limit
        self.round_limit = round_limit
        self.time_limit = time_limit
        self.boolean_limit = boolean_limit
        self.random_method = random_method

        # set a counter
        self.round_counter = 0

        # handle tuple conversion to discrete values
        self.p = self._param_input_conversion()

        # create list of list from the params dictionary
        self._params_temp = [list(self.p[key]) for key in self.param_keys]

        # establish max dimensions
        self.dimensions = np.prod([len(l) for l in self._params_temp])

        # apply all the set limits
        self.param_index = self._param_apply_limits()

        # create the parameter space
        self.param_space = self._param_space_creation()

        # handle the boolean limits separately
        if self.boolean_limit is not None:
            index = self._convert_lambda(self.boolean_limit)(self.param_space)
            self.param_space = self.param_space[index]

        # reset index
        self.param_index = list(range(len(self.param_index)))

    def _param_input_conversion(self):

        '''Parameters may be input as lists of single or
        multiple values (discrete values) or tuples
        (range of values). This helper checks the format of
        each input and handles it accordingly.'''

        out = {}

        # go through each parameter type
        for param in self.param_keys:

            # deal with range (tuple) values
            if isinstance(self.params[param], tuple):
                out[param] = self._param_range_expansion(self.params[param])

            # deal with range (list) values
            elif isinstance(self.params[param], list):
                out[param] = self.params[param]

        return out

    def _param_apply_limits(self):

        from talos.reducers.sample_reducer import sample_reducer

        if self.boolean_limit is not None:
            # NOTE: this is handled in __init__
            pass

        # a time limit is set
        if self.time_limit is not None:
            # NOTE: this is handled in _time_left
            pass

        # a fractional limit is set
        if self.fraction_limit is not None:
            return sample_reducer(self.fraction_limit,
                                  self.dimensions,
                                  self.random_method)

        # a round limit is set
        if self.round_limit is not None:
            return sample_reducer(self.round_limit,
                                  self.dimensions,
                                  self.random_method)

        # no limits are set
        return list(range(self.dimensions))

    def _param_range_expansion(self, param_values):

        '''Expands a range (tuple) input into discrete
        values. Helper for _param_input_conversion.
        Expects to have a input as (start, end, steps).
        '''

        start = param_values[0]
        end = param_values[1]
        steps = param_values[2]

        out = np.arange(start, end, (end - start) / steps, dtype=float)

        # inputs are all ints
        if isinstance(start, int) and isinstance(end, int):
            out = out.astype(int)
            out = np.unique(out)

        return out

    def _param_space_creation(self):

        '''Expand params dictionary to permutations

        Takes the input params dictionary and expands it to
        actual parameter permutations for the experiment.
        '''

        # handle the cases where parameter space is still large
        if len(self.param_index) > 100000:

            final_grid = list(it.product(*self._params_temp))
            out = np.array(final_grid, dtype='object')

        # handle the cases where parameter space is already smaller
        else:
            final_grid = []
            for i in self.param_index:
                p = []
                for l in reversed(self._params_temp):
                    i, s = divmod(int(i), len(l))
                    p.insert(0, l[s])
                final_grid.append(tuple(p))

            out = np.array(final_grid, dtype='object')

        return out

    def _check_time_limit(self):

        if self.time_limit is None:
            return True

        stop = datetime.strptime(self.time_limit, "%Y-%m-%d %H:%M")

        return stop > datetime.now()

    def round_parameters(self):

        # permutations remain in index
        if len(self.param_index) > 0:

            # time limit has not been met yet
            if self._check_time_limit():
                self.round_counter += 1

                # get current index
                index = self.param_index.pop(0)

                # get the values based on the index
                values = self.param_space[index]
                round_parameters = self._round_parameters_todict(values)

                # pass the parameters to Scan
                return round_parameters

        # the experiment is finished
        return False

    def _round_parameters_todict(self, values):

        round_parameters = {}

        for i, key in enumerate(self.param_keys):
            round_parameters[key] = values[i]

        return round_parameters

    def _convert_lambda(self, fn):

        '''Converts a lambda function into a format
        where parameter labels are changed to the column
        indexes in parameter space.'''

        # get the source code for the lambda function
        fn_string = inspect.getsource(fn)
        fn_string = fn_string.replace('"', '\'')

        # look for column/label names
        for i, name in enumerate(self.param_keys):
            index = ':,' + str(i)
            fn_string = fn_string.replace(name, index)

        # cleanup the string
        fn_string = fn_string.split('lambda')[1]
        fn_string = fn_string.replace('[\':', '[:')
        fn_string = fn_string.replace('\']', ']')
        fn_string = 'lambda ' + fn_string

        # pass it back as a function
        return eval(fn_string)

    def remove_is_not(self, label, value):

        '''Removes baesd on exact match but reversed'''

        col = self.param_keys.index(label)
        drop = np.where(self.param_space[:, col] != value)[0].tolist()
        self.param_index = [x for x in self.param_index if x not in drop]

    def remove_is(self, label, value):

        '''Removes based on exact match'''

        col = self.param_keys.index(label)
        drop = np.where(self.param_space[:, col] == value)[0].tolist()
        self.param_index = [x for x in self.param_index if x not in drop]

    def remove_ge(self, label, value):

        '''Removes based on greater-or-equal'''

        col = self.param_keys.index(label)
        drop = np.where(self.param_space[:, col] >= value)[0].tolist()
        self.param_index = [x for x in self.param_index if x not in drop]

    def remove_le(self, label, value):

        '''Removes based on lesser-or-equal'''

        col = self.param_keys.index(label)
        drop = np.where(self.param_space[:, col] <= value)[0].tolist()
        self.param_index = [x for x in self.param_index if x not in drop]

    def remove_lambda(self, function):

        '''Removes based on a lambda function'''

        index = self._convert_lambda(function)(self.param_space)
        self.param_space = self.param_space[index]
        self.param_index = list(range(len(self.param_space)))



import numpy as np


def network_shape(params, last_neuron):

    '''Provides the ability to include network shape in experiments. If params
    dictionary for the round contains float value for params['shapes'] then
    a linear contraction towards the last_neuron value. The higher the value,
    the fewer layers it takes to reach lesser than last_neuron.

    Supports three inbuilt shapes 'brick', 'funnel', and 'triangle'.


    params : dict
         Scan() params for a single roundself.
    last_neuron : int
         Number of neurons on the output layer in the Keras model.
    '''

    layers = params['hidden_layers']
    shape = params['shapes']
    first_neuron = params['first_neuron']
    out = []
    n = first_neuron

    # the case where hidden_layers is zero
    if layers == 0:
        return [0]

    # the cases where an angle is applied
    if isinstance(shape, float):

        for i in range(layers):

            n *= 1 - shape

            if n > last_neuron:
                out.append(int(n))
            else:
                out.append(last_neuron)

    # the case where a rectantular shape is used
    elif shape == 'brick':
        out = [first_neuron] * layers

    elif shape == 'funnel':
        for i in range(layers + 1):
            n -= int((first_neuron - last_neuron) / layers)
            out.append(n)
        out.pop(-1)

    elif shape == 'triangle':
        out = np.linspace(first_neuron,
                          last_neuron,
                          layers+2,
                          dtype=int).tolist()
        out.pop(0)
        out.pop(-1)
        out.reverse()

    return out



import numpy as np
from keras.optimizers import Adam, Nadam, Adadelta, SGD


loss = {'binary': ['binary_crossentropy', 'logcosh'],
        'multi_class': ['sparse_categorical_crossentropy'],
        'multi_label': ['categorical_crossentropy'],
        'continuous': ['mae']}

last_activation = {'binary': ['sigmoid'],
                   'multi_class': ['softmax'],
                   'multi_label': ['softmax'],
                   'continuous': [None]}


class AutoParams:

    def __init__(self,
                 params=None,
                 task='binary',
                 replace=True,
                 auto=True,
                 network=True):

        '''A facility for generating or appending params dictionary.

        params : dict or None
        task : str
             'binary', 'multi_class', 'multi_label', or 'continuous'
        replace : bool
             Replace current dictionary entries with new ones.
        auto : bool
             Automatically generate or append params dictionary with
             all available parameters.
        network : bool
             Adds several network architectures as parameters. This is to be
             used as an input together with KerasModel(). If False then only
             'dense' will be added.
        '''

        self._task = task
        self._replace = replace
        self._network = network

        if params is None:
            self.params = {}
        else:
            self.params = params
        if auto:
            self._automated()
            self.resample_params(4)

    def _automated(self, shapes='fixed'):

        '''Automatically generate a comprehensive
        parameter dict to be used in Scan()

        shapes : string
            Either 'fixed' or 'sloped'

        '''

        if shapes == 'fixed':
            self.shapes()
        else:
            self.shapes_slope()
        self.layers()
        self.dropout()
        self.optimizers()
        self.activations()
        self.neurons()
        self.losses()
        self.batch_size()
        self.epochs()
        self.kernel_initializers()
        self.lr()
        if self._network:
            self.networks()
        else:
            self.params['network'] = ['dense']
        self.last_activations()

    def shapes(self, shapes='auto'):

        '''Uses triangle, funnel, and brick shapes.'''

        if shapes == 'auto':
            self._append_params('shapes', ['triangle', 'funnel', 'brick'])
        else:
            self._append_params('shapes', shapes)

    def shapes_slope(self, min=0, max=.6, steps=.1):

        '''Uses a single decimal float for values below 0.5 to
        reduce the width of the following layer.'''

        self._append_params('shapes', np.arange(min, max, steps).tolist())

    def layers(self, min_layers=0, max_layers=6, steps=1):

        self._append_params('hidden_layers',
                            list(range(min_layers, max_layers, steps)))

    def dropout(self, min=0, max=.85, steps=0.1):

        self._append_params('dropout',
                            np.round(np.arange(min, max, steps), 2).tolist())

    def optimizers(self, optimizers='auto'):

        '''If `optimizers='auto'` then optimizers will be picked based on
        automatically. Otherwise input a list with one or
        more optimizers will be used.
        '''

        if optimizers == 'auto':
            self._append_params('optimizer', [Adam, Nadam, Adadelta, SGD])
        else:
            self._append_params('optimizer', optimizers)

    def activations(self, activations='auto'):

        '''If `activations='auto'` then activations will be picked based on
        automatically. Otherwise input a list with one or
        more activations will be used.
        '''

        if activations == 'auto':
            activations = ['relu', 'elu']

        self._append_params('activation', activations)

    def losses(self, losses='auto'):

        '''If `losses='auto'` then losses will be picked based on
        `AutoParam()` argument `task`. Otherwise input a list with one or
        more losses will be used.
        '''

        if losses == 'auto':
            self._append_params('losses', loss[self._task])
        else:
            self._append_params('losses', losses)

    def neurons(self, min=8, max=None, steps=None):

        '''`max` and `steps` has to be either `None` or
        integer value at the same time.'''

        if max is None and steps is None:
            values = [int(np.exp2(i)) for i in range(3, 11)]
        else:
            values = list(range(min, max, steps))

        self._append_params('first_neuron', values)

    def batch_size(self, min=8, max=None, steps=None):

        '''`max` and `steps` has to be either `None` or
        integer value at the same time.'''

        if max is None and steps is None:
            values = [int(np.exp2(i/2)) for i in range(3, 15)]
        else:
            values = list(range(min, max, steps))

        self._append_params('batch_size', values)

    def epochs(self, min=50, max=None, steps=None):

        '''`max` and `steps` has to be either `None` or
        integer value at the same time.'''

        if max is None and steps is None:
            values = [int(np.exp2(i/2))+50 for i in range(3, 15)]
        else:
            values = list(range(min, max, steps))

        self._append_params('epochs', values)

    def kernel_initializers(self, kernel_inits='auto'):

        '''
        kernel_inits | list | one or more kernel initializers
        '''

        if kernel_inits == 'auto':
            self._append_params('kernel_initializer',
                                ['uniform', 'normal', 'he_normal',
                                 'he_uniform', 'lecun_normal',
                                 'glorot_uniform', 'glorot_normal',
                                 'random_uniform', 'random_normal'])
        else:
            self._append_params('kernel_initializer', kernel_inits)

    def lr(self, learning_rates='auto'):

        '''If `learning_rates='auto'` then a very wide range of learning rates
        will be added. Otherwise a list with one or more learning rates
        is used.

        NOTE: talos.utils.lr_normalizer should be used if more than one optimizer
        is used in the experiment
        '''

        if learning_rates == 'auto':

            a = np.round(np.arange(0.01, 0.2, 0.02), 3).tolist()
            b = np.round(np.arange(0, 1, 0.2), 2).tolist()
            c = list(range(0, 11))

            self._append_params('lr', a + b + c)

        else:

            self._append_params('lr', learning_rates)

    def networks(self, networks='auto'):

        '''If `network='auto'` then dense, simplernn, lstm, conv1d, and
        bidirectional_lstm are added. Otherwise a list with one or more
        network architectures is used.
        '''

        if networks == 'auto':
            self._append_params('network', ['dense',
                                            'simplernn',
                                            'lstm',
                                            'bidirectional_lstm',
                                            'conv1d'])
        else:
            self._append_params('network', networks)

    def last_activations(self, last_activations='auto'):

        '''If `last_activations='auto'` then activations will be picked
        automatically based on `AutoParams` property `task`.
        Otherwise input a list with one or more activations will be used.
        '''

        if last_activations == 'auto':
            self._append_params('last_activation', last_activation[self._task])
        else:
            self._append_params('last_activation', last_activations)

    def resample_params(self, n):

        '''Resamples params dictionary so that `n` values are present for each
        parameter.'''

        from wrangle import dic_resample_values

        self.params = dic_resample_values(self.params, n)

    def _append_params(self, label, values):

        if self._replace is False:
            try:
                self.params[label]
            except KeyError:
                self.params[label] = values

        else:
            self.params[label] = values



def scan_run(self):

    '''The high-level management of the scan procedures
    onwards from preparation. Manages round_run()'''

    from tqdm import tqdm

    from .scan_prepare import scan_prepare
    self = scan_prepare(self)

    # initiate the progress bar
    self.pbar = tqdm(total=len(self.param_object.param_index),
                     disable=self.disable_progress_bar)

    # the main cycle of the experiment
    while True:

        # get the parameters
        self.round_params = self.param_object.round_parameters()

        # break when there is no more permutations left
        if self.round_params is False:
            break
        # otherwise proceed with next permutation
        from .scan_round import scan_round
        self = scan_round(self)
        self.pbar.update(1)

    # close progress bar before finishing
    self.pbar.close()

    # finish
    from ..logging.logging_finish import logging_finish
    self = logging_finish(self)

    from .scan_finish import scan_finish
    self = scan_finish(self)



def reduce_run(self):

    '''The process run script for reduce
    procedures; takes care of everything
    related with reduction. When new
    reduction methods are added, they need
    to be added as options here.

    To add new reducers, create a file in /reducers
    which is where this file is located. In that file,
    take as input self from Scan() and give as output
    either False, which does nothing, or a tuple of
    'value' and 'label' where value is a parameter
    value and label is parameter name. For example
    batch_size and 128. Then add a reference to
    reduce_run.py and make sure that you process
    the self.param_object.param_index there before
    wrapping up.

    '''

    from .correlation import correlation
    from .forrest import forrest
    from .trees import trees
    from .gamify import gamify

    from .local_strategy import local_strategy
    from .limit_by_metric import limit_by_metric

    # check if performance target is met
    if self.performance_target is not None:
        status = limit_by_metric(self)

        # handle the case where performance target is met
        if status == True:
            self.param_object.param_index = []
            print("Target %.3f have been met." % self.performance_target[1])

    # stop here if no reduction method is set
    if self.reduction_method is None:
        return self

    # setup what's required for updating progress bar
    left = (self.param_object.round_counter + 1)
    right = self.reduction_interval
    len_before_reduce = len(self.param_object.param_index)

    # check if monte carlo can do something
    if self.reduction_method == 'gamify':
        self = gamify(self)

    # apply window based reducers
    if left % right == 0:

        # check if correlation reducer can do something
        if self.reduction_method in ['pearson', 'kendall', 'spearman']:
            self = correlation(self, self.reduction_method)

        # check if correlation reducer can do something
        if self.reduction_method == 'correlation':
            self = correlation(self, 'spearman')

        # check if random forrest can do something
        if self.reduction_method == 'forrest':
            self = forrest(self)

        # check if random forrest can do something
        if self.reduction_method == 'trees':
            self = trees(self)

        if self.reduction_method == 'local_strategy':
            self = local_strategy(self)

        # finish up by updating progress bar
        total_reduced = len_before_reduce - len(self.param_object.param_index)
        total_reduced = max(0, total_reduced)
        self.pbar.update(total_reduced)

        if total_reduced > 0:
            # print out the the status
            drop_share = total_reduced / len_before_reduce * 100
            print("Total %.1f%% permutations reduced" % drop_share)

    return self



def ingest_model(self):

    '''Ingests the model that is input by the user
    through Scan() model paramater.'''

    return self.model(self.x_train,
                      self.y_train,
                      self.x_val,
                      self.y_val,
                      self.round_params)



def limit_by_metric(self):

    '''Takes as input metric, threshold, and loss and
    and returs a True if metric threshold have been
    met and False if not.

    USE: space.check_metric(model_history)
    '''

    metric = self.performance_target[0]
    threshold = self.performance_target[1]
    loss = self.performance_target[2]

    if loss is True:
        return self.model_history.history[metric][-1] <= threshold
    elif loss is False:
        return self.model_history.history[metric][-1] >= threshold



def cols_to_multilabel(self):

    '''Utility function for correlation and other reducers
    that require transforming hyperparameter values into
    multilabel values before applying the reduction strategy.'''

    import wrangle
    import pandas as pd

    # read in the experiment log
    data = pd.read_csv(self._experiment_log)

    # apply recuction window
    data = data.tail(self.reduction_window)

    # drop all other metric columns except reduction_metric
    data = data[[self.reduction_metric] + self._param_dict_keys]

    # convert all hyperparameter columns to multi label columns
    for col in data.iloc[:, 1:].columns:

        # get the dtype of the column data
        col_dtype = data[col].dtype

        # parse column name to contain label, value and dtype
        data = wrangle.col_to_multilabel(data,
                                         col,
                                         extended_colname=True,
                                         extended_separator='~' + str(col_dtype) + '~')

    return data
